@@ -1,7 +1,11 @@
 "pymarc marcxml file."
 
+import logging
 from xml.sax import make_parser
 from xml.sax.handler import ContentHandler, feature_namespaces
+import unicodedata
+
+import six
 
 try:
     import xml.etree.ElementTree as ET  # builtin in Python 2.5
@@ -23,13 +27,14 @@ class XmlHandler(ContentHandler):
     records elsewhere (like to a rdbms) without having to store 
     them all in memory.
     """
-    def __init__(self, strict=False):
+    def __init__(self, strict=False, normalize_form=None):
         self.records = []
         self._record = None
         self._field = None
         self._subfield_code = None
         self._text = []
         self._strict = strict
+        self.normalize_form = normalize_form
 
     def startElementNS(self, name, qname, attrs):
         if self._strict and name[0] != MARC_XML_NS:
@@ -45,8 +50,8 @@ class XmlHandler(ContentHandler):
             self._field = Field(tag)
         elif element == 'datafield':
             tag = attrs.getValue((None, u'tag'))
-            ind1 = attrs.getValue((None, u'ind1'))
-            ind2 = attrs.getValue((None, u'ind2'))
+            ind1 = attrs.get((None, u'ind1'), u' ')
+            ind2 = attrs.get((None, u'ind2'), u' ')
             self._field = Field(tag, [ind1, ind2])
         elif element == 'subfield':
             self._subfield_code = attrs[(None, 'code')]
@@ -56,7 +61,10 @@ class XmlHandler(ContentHandler):
             return
 
         element = name[1]
-        text = u''.join(self._text)
+        if self.normalize_form is not None:
+            text = unicodedata.normalize(self.normalize_form, u''.join(self._text))
+        else:
+            text = u''.join(self._text)
 
         if element == 'record':
             self.process_record(self._record)
@@ -98,7 +106,7 @@ def map_xml(function, *files):
     parsed the function will get called with the extracted record
 
     def do_it(r):
-      print r
+      print(r)
 
     map_xml(do_it, 'marc.xml')
     """
@@ -107,13 +115,15 @@ def map_xml(function, *files):
     for xml_file in files:
         parse_xml(xml_file, handler)
 
-def parse_xml_to_array(xml_file, strict=False):
+def parse_xml_to_array(xml_file, strict=False, normalize_form=None):
     """
     parse an xml file and return the records as an array. If you would
     like the parser to explicitly check the namespaces for the MARCSlim
     namespace use the strict=True option.
+    Valid values for normalize_form are 'NFC', 'NFKC', 'NFD', and 'NFKD'. See
+    unicodedata.normalize info.
     """
-    handler = XmlHandler(strict)
+    handler = XmlHandler(strict, normalize_form)
     parse_xml(xml_file, handler)
     return handler.records
 
@@ -128,7 +138,7 @@ def record_to_xml(record, quiet=False, namespace=False):
     # TODO: maybe should set g0 and g1 appropriately using 066 $a and $b?
     marc8 = MARC8ToUnicode(quiet=quiet)
     def translate(data):
-        if type(data) == unicode: 
+        if type(data) == six.text_type: 
             return data
         else: 
             return marc8.translate(data)
