@@ -1,16 +1,32 @@
 import re
+import logging
+
+from six import Iterator
+from six.moves import zip_longest as izip_longest
 
 from pymarc.exceptions import BaseAddressInvalid, RecordLeaderInvalid, \
         BaseAddressNotFound, RecordDirectoryInvalid, NoFieldsFound, \
         FieldNotFound
 from pymarc.constants import LEADER_LEN, DIRECTORY_ENTRY_LEN, END_OF_RECORD
 from pymarc.field import Field, SUBFIELD_INDICATOR, END_OF_FIELD, \
-        map_marc8_field
+        map_marc8_field, RawField
 from pymarc.marc8 import marc8_to_unicode
 
-isbn_regex = re.compile(r'([0-9\-]+)')
+try:
+    # the json module was included in the stdlib in python 2.6
+    # http://docs.python.org/library/json.html
+    import json
+except ImportError:
+    # simplejson 2.0.9 is available for python 2.4+
+    # http://pypi.python.org/pypi/simplejson/2.0.9
+    # simplejson 1.7.3 is available for python 2.3+
+    # http://pypi.python.org/pypi/simplejson/1.7.3
+    import simplejson as json
 
-class Record(object):
+
+isbn_regex = re.compile(r'([0-9\-xX]+)')
+
+class Record(Iterator):
     """
     A class for representing a MARC record. Each Record object is made up of
     multiple Field objects. You'll probably want to look at the docs for Field
@@ -41,16 +57,19 @@ class Record(object):
     MARC records in a file.  
     """
 
-    def __init__(self, data='', to_unicode=False, force_utf8=False,
+    def __init__(self, data='', to_unicode=True, force_utf8=False,
         hide_utf8_warnings=False, utf8_handling='strict'):
         self.leader = (' '*10) + '22' + (' '*8) + '4500'
         self.fields = list()
         self.pos = 0
+        self.force_utf8 = force_utf8
         if len(data) > 0:
             self.decode_marc(data, to_unicode=to_unicode,
                              force_utf8=force_utf8,
                              hide_utf8_warnings=hide_utf8_warnings,
                              utf8_handling=utf8_handling)
+        elif force_utf8:
+            self.leader = self.leader[0:9] + 'a' + self.leader[10:]
 
     def __str__(self):
         """
@@ -76,11 +95,21 @@ class Record(object):
             return fields[0]
         return None
 
+    def __contains__(self, tag):
+        """
+        Allows a shorthand test of tag membership:
+
+            '245' in record
+
+        """
+        fields = self.get_fields(tag)
+        return len(fields) > 0
+
     def __iter__(self):
         self.__pos = 0
         return self
 
-    def next(self):
+    def __next__(self):
         if self.__pos >= len(self.fields):
             raise StopIteration
         self.__pos += 1 
@@ -93,6 +122,56 @@ class Record(object):
         """
         self.fields.extend(fields)
 
+    def add_grouped_field(self, *fields):
+        """
+        add_grouped_field() will add pymarc.Field objects to a Record object,
+        attempting to maintain a loose numeric order per the MARC standard for
+        "Organization of the record" (http://www.loc.gov/marc/96principl.html)
+        Optionally you can pass in multiple fields.
+        """
+        for f in fields:
+            if len(self.fields) == 0 or not f.tag.isdigit():
+                self.fields.append(f)
+                continue 
+            self._sort_fields(f, 'grouped')
+
+    def add_ordered_field(self, *fields):
+        """
+        add_ordered_field() will add pymarc.Field objects to a Record object,
+        attempting to maintain a strict numeric order.
+        Optionally you can pass in multiple fields.
+        """
+        for f in fields:
+            if len(self.fields) == 0 or not f.tag.isdigit():
+                self.fields.append(f)
+                continue 
+            self._sort_fields(f, 'ordered')
+
+    def _sort_fields(self, field, mode):
+        if mode == 'grouped':
+            tag = int(field.tag[0])
+        else:
+            tag = int(field.tag)
+
+        i, last_tag = 0, 0
+        for selff in self.fields:
+            i += 1
+            if not selff.tag.isdigit():
+                self.fields.insert(i - 1, field)
+                break
+
+            if mode == 'grouped':
+                last_tag = int(selff.tag[0])
+            else:
+                last_tag = int(selff.tag)
+
+            if last_tag > tag:
+                self.fields.insert(i - 1, field)
+                break
+            if len(self.fields) == i:
+                self.fields.append(field)
+                break
+
     def remove_field(self, *fields):
         """
         remove_field() will remove one or more pymarc.Field objects from
@@ -117,7 +196,7 @@ class Record(object):
 
             subjects = record.get_fields('600', '610', '650') 
 
-        If no tag is passed in to fields() a list of all the fields will be 
+        If no tag is passed in to get_fields() a list of all the fields will be 
         returned.
         """
         if (len(args) == 0):
@@ -125,7 +204,7 @@ class Record(object):
 
         return [f for f in self.fields if f.tag in args]
 
-    def decode_marc(self, marc, to_unicode=False, force_utf8=False,
+    def decode_marc(self, marc, to_unicode=True, force_utf8=False,
         hide_utf8_warnings=False, utf8_handling='strict'):
         """
         decode_marc() accepts a MARC record in transmission format as a
@@ -135,10 +214,15 @@ class Record(object):
 
         """
         # extract record leader
-        self.leader = marc[0:LEADER_LEN]
+        self.leader = marc[0:LEADER_LEN].decode('ascii')
         if len(self.leader) != LEADER_LEN: 
             raise RecordLeaderInvalid
 
+        if self.leader[9] == 'a' or self.force_utf8:
+            encoding = 'utf-8'
+        else:
+            encoding = 'iso8859-1'
+
         # extract the byte offset where the record data starts
         base_address = int(marc[12:17])
         if base_address <= 0: 
@@ -148,7 +232,7 @@ class Record(object):
 
         # extract directory, base_address-1 is used since the 
         # director ends with an END_OF_FIELD byte
-        directory = marc[LEADER_LEN:base_address-1]
+        directory = marc[LEADER_LEN:base_address-1].decode('ascii')
 
         # determine the number of fields in record
         if len(directory) % DIRECTORY_ENTRY_LEN != 0:
@@ -169,16 +253,44 @@ class Record(object):
 
             # assume controlfields are numeric; replicates ruby-marc behavior 
             if entry_tag < '010' and entry_tag.isdigit():
-                field = Field(tag=entry_tag, data=entry_data)
+                if to_unicode:
+                    field = Field(tag=entry_tag, data=entry_data.decode(encoding))
+                else:
+                    field = RawField(tag=entry_tag, data=entry_data)
             else:
                 subfields = list()
-                subs = entry_data.split(SUBFIELD_INDICATOR)
-                first_indicator = subs[0][0]
-                second_indicator = subs[0][1]
+                subs = entry_data.split(SUBFIELD_INDICATOR.encode('ascii'))
+
+                # The MARC spec requires there to be two indicators in a
+                # field. However experience in the wild has shown that
+                # indicators are sometimes missing, and sometimes there
+                # are too many. Rather than throwing an exception because
+                # we can't find what we want and rejecting the field, or
+                # barfing on the whole record we'll try to use what we can
+                # find. This means missing indicators will be recorded as
+                # blank spaces, and any more than 2 are dropped on the floor.
+
+                first_indicator = second_indicator = ' '
+                subs[0] = subs[0].decode('ascii')
+                if len(subs[0]) == 0:
+                    logging.warning("missing indicators: %s", entry_data)
+                    first_indicator = second_indicator = ' '
+                elif len(subs[0]) == 1:
+                    logging.warning("only 1 indicator found: %s", entry_data)
+                    first_indicator = subs[0][0]
+                    second_indicator = ' '
+                elif len(subs[0]) > 2:
+                    logging.warning("more than 2 indicators found: %s", entry_data)
+                    first_indicator = subs[0][0]
+                    second_indicator = subs[0][1]
+                else:
+                    first_indicator = subs[0][0]
+                    second_indicator = subs[0][1]
+
                 for subfield in subs[1:]:
-                    if len(subfield) == 0: 
+                    if len(subfield) == 0:
                         continue
-                    code = subfield[0]
+                    code = subfield[0:1].decode('ascii')
                     data = subfield[1:]
 
                     if to_unicode:
@@ -188,12 +300,18 @@ class Record(object):
                             data = marc8_to_unicode(data, hide_utf8_warnings)
                     subfields.append(code)
                     subfields.append(data)
-                field = Field( 
-                    tag = entry_tag, 
-                    indicators = [first_indicator, second_indicator], 
-                    subfields = subfields,
-                )
-
+                if to_unicode:
+                    field = Field( 
+                        tag = entry_tag, 
+                        indicators = [first_indicator, second_indicator], 
+                        subfields = subfields,
+                    )
+                else:
+                    field = RawField(
+                        tag = entry_tag, 
+                        indicators = [first_indicator, second_indicator], 
+                        subfields = subfields,
+                    )
             self.add_field(field)
             field_count += 1
 
@@ -204,30 +322,35 @@ class Record(object):
         """
         returns the record serialized as MARC21
         """
-        fields = ''
-        directory = '' 
+        fields = b''
+        directory = b'' 
         offset = 0
 
         # build the directory
         # each element of the directory includes the tag, the byte length of 
         # the field and the offset from the base address where the field data
         # can be found
+        if self.leader[9] == 'a' or self.force_utf8:
+            encoding = 'utf-8'
+        else:
+            encoding = 'iso8859-1'
+
         for field in self.fields:
-            field_data = field.as_marc().encode('utf-8')
+            field_data = field.as_marc(encoding=encoding)
             fields += field_data
             if field.tag.isdigit():
-                directory += '%03d' % int(field.tag)
+                directory += ('%03d' % int(field.tag)).encode(encoding)
             else:
-                directory += '%03s' % field.tag
-            directory += '%04d%05d' % (len(field_data), offset)
+                directory += ('%03s' % field.tag).encode(encoding)
+            directory += ('%04d%05d' % (len(field_data), offset)).encode(encoding)
     
             offset += len(field_data)
 
         # directory ends with an end of field
-        directory += END_OF_FIELD
+        directory += END_OF_FIELD.encode(encoding)
 
         # field data ends with an end of record
-        fields += END_OF_RECORD
+        fields += END_OF_RECORD.encode(encoding)
 
         # the base address where the directory ends and the field data begins
         base_address = LEADER_LEN + len(directory)
@@ -237,15 +360,42 @@ class Record(object):
 
         # update the leader with the current record length and base address
         # the lengths are fixed width and zero padded
-        self.leader = '%05d%s%05d%s' % \
+        strleader = '%05d%s%05d%s' % \
             (record_length, self.leader[5:12], base_address, self.leader[17:])
+        self.leader = strleader.encode(encoding)
         
-        # return the encoded record
-        return self.leader + directory + fields 
+        return self.leader + directory + fields
 
     # alias for backwards compatability
     as_marc21 = as_marc
 
+    def as_dict(self):
+        """
+        Turn a MARC record into a dictionary, which is used for ``as_json``.
+        """
+        record = {}
+        record['leader'] = self.leader
+        record['fields'] = []
+        for field in self:
+            if field.is_control_field():
+                record['fields'].append({field.tag: field.data})
+            else:
+                fd = {}
+                fd['subfields'] = []
+                fd['ind1'] = field.indicator1
+                fd['ind2'] = field.indicator2
+                for tag, value in izip_longest(*[iter(field.subfields)] * 2):
+                    fd['subfields'].append({tag: value})
+                record['fields'].append({field.tag: fd})
+        return record  # as dict
+
+    def as_json(self, **kwargs):
+        """
+        Serialize a record as JSON according to
+        http://dilettantes.code4lib.org/blog/2010/09/a-proposal-to-serialize-marc-in-json/
+        """
+        return json.dumps(self.as_dict(), **kwargs)
+
     def title(self):
         """
         Returns the title of the record (245 $a an $b).
@@ -264,7 +414,8 @@ class Record(object):
     def isbn(self):
         """
         Returns the first ISBN in the record or None if one is not
-        present. The returned ISBN will be all numberic; so dashes and 
+        present. The returned ISBN will be all numeric, except for an
+        x/X which may occur in the checksum position.  Dashes and 
         extraneous information will be automatically removed. If you need 
         this information you'll want to look directly at the 020 field, 
         e.g. record['020']['a']
@@ -294,7 +445,15 @@ class Record(object):
         elif self['240']:
             return self['240'].format_field()
         return None
-
+    
+    def series(self):
+        """
+        Note: 490 supersedes the 440 series statement which was both 
+        series statement and added entry. 8XX fields are added entries.
+        """
+        serieslist = self.get_fields('440', '490', '800', '810', '811', '830')
+        return serieslist
+        
     def subjects(self):
         """
         Note: Fields 690-699 are considered "local" added entry fields but
@@ -342,11 +501,17 @@ class Record(object):
     def publisher(self):
         if self['260']:
             return self['260']['b']
+        if self['264']:
+            return self['264']['b']
+
         return None
 
     def pubyear(self):
         if self['260']:
             return self['260']['c']
+        elif self['264']:
+            return self['264']['c']
+
         return None
 
 def map_marc8_record(r):
